grammar CalangoGrammar;

options {
  language = Java;
  output = AST; 
  ASTLabelType = CommonTree;
}
 
tokens {
  PARAMETRO;
  PARAMETROS;
  PARAMETROS_CHAMADA;
  CHAMADA;
  DECL_VARIAVEL;
  REFERENCIA = 'ref';
  PARAM_VETOR = '[]';
  PARAM_MATRIZ = '[][]';
  NEGACAO_MATEMATICA;
  ENQUANTO_STM;
  FUNCAO = 'funcao';
  FIM_FUNCAO = 'fimFuncao';
  PROCEDIMENTO = 'procedimento';
  FIM_PROCEDIMENTO = 'fimProcedimento';
  PRINCIPAL;
  INTERROMPA = 'interrompa';
  ATRIBUICAO;
  SE = 'se';
  ESCOLHA = 'escolha';
  CASO = 'caso';
  OUTROCASO = 'outroCaso';
  PARA = 'para';
  FACA = 'faca';
  ENQUANTO = 'enquanto';
  LEIA = 'leia';
  LEIA_CARACTER = 'leiaCaracter';
  ESCREVA = 'escreva';
  ESCREVAL = 'escreval';
  RETORNA = 'retorna';
  TIPO_INTEIRO = 'inteiro';
  TIPO_REAL = 'real';
  TIPO_CARACTER = 'caracter';
  TIPO_TEXTO = 'texto';
  TIPO_LOGICO = 'logico';
  CONCATENACAO = ',';
  E = 'e';
  OU = 'ou';
  IGUAL = '==';
  DIFERENTE = '!=';
  MENOR = '<';
  MENOR_IGUAL = '<=';
  MAIOR_IGUAL = '>=';
  MAIOR = '>';
  ADICAO = '+';
  SUBTRACAO = '-';
  MULTIPLICACAO = '*';
  DIVISAO = '/';
  DIVISAO_INT = '\\';
  MOD = 'mod';
  NEGACAO_LOGICA = 'nao';  
  VETOR;
  DECL_ATRIBUICAO;
  TIPO_IDENTIFICADOR;
  SE_STMS;
  SENAO_STMS;
  CASOS;
  STATEMENTS;
  ALGORITMO;
  TERMO_VETOR;
  CORPO;
  COMPARA_TEXTO;
  TAMANHO_TEXTO;
  COPIA;
  ABS;
  EXP;
  PI;
  RQUAD;
  ALEAT;
  MAIUSC;
  MINUSC;
  MAIUSC_CHAR;
  MINUSC_CHAR;
  ASCII_CHAR;
  CHAR_ASCII;
  LIMPA_TELA = 'limpaTela';
  SENAO_SE;
  SENAO;
  IDENT;
  FORMATA_REAL;
  FORMATA_REAL_TOTAL;
  FORMATA_REAL_DECIMAL;
}

@header {
  package br.ucb.calango.gramatica;
  import br.ucb.calango.arvore.CalangoTree;
  import br.ucb.calango.util.*;
  import br.ucb.calango.api.publica.CalangoAPI;
  import br.ucb.calango.exceptions.parser.ComandoForaDoContextoException;
} 

@lexer::header {  
  package br.ucb.calango.gramatica;
  import br.ucb.calango.util.*;
  import br.ucb.calango.api.publica.CalangoAPI;
}

@lexer::members {
  private boolean errors = false;
  public void emitErrorMessage(String msg) {
    errors = true;
    CalangoAPI.printErro(msg);
  }
  public String getErrorMessage(RecognitionException e, String[] tokenNames) {
    return MensagensUtil.getMessage(e.getClass(), ErrosUtil.getMessageParams(e));
  }
  public String getErrorHeader(RecognitionException e) {
    return "Linha " + e.line + " >>";
  }
  public boolean hasErrors(){
    return errors;
  }
}

@members {
  private boolean errors = false;
  public static int previousLineNumber = 0;
  
  public void emitErrorMessage(String msg) {
    errors = true;
    CalangoAPI.printErro(msg);
  }
  
  public String getErrorMessage(RecognitionException e, String[] tokenNames) {
    if(e instanceof MissingTokenException && ((MissingTokenException)e).expecting == CalangoGrammarParser.EOF){
      // Comando fora do contexto
      e = new ComandoForaDoContextoException();
    } 
    return MensagensUtil.getMessage(e.getClass(), ErrosUtil.getMessageParams(e)); 
  }
  
  public String getErrorHeader(RecognitionException e) {
    int lineNum;
    // Se for um MissingToken e esse token for ';', recupera o número de linha do token anterior.
    if(e instanceof MissingTokenException && 
       ((MissingTokenException)e).expecting > 0 &&
       tokenNames[((MissingTokenException)e).expecting].equals("';'")){
      lineNum = input.LT(-1).getLine();
    } else {
      lineNum = e.line;
    }
    return "Linha " + lineNum + " >>"; // TODO remover string escrita diretamente aqui
  }
  
  public boolean hasErrors(){
    return errors;
  }
  
  private CommonToken createToken(String text, int type, int line) {
    CommonToken t = new CommonToken(type, text);
    t.setLine(line);
    return t;
  }
}
 
algoritmo
    : 'algoritmo' IDENTIFICADOR ';' 
      principal 
      funcao_ou_procedimento*
      EOF
      -> ^(ALGORITMO funcao_ou_procedimento* principal EOF) 
    ;  
    
funcao_ou_procedimento 
    : (funcao | procedimento)
    ;
       
funcao
    : FUNCAO tipo_dado IDENTIFICADOR '(' decl_parametros? ')'
      corpo
      FIM_FUNCAO 
      
      -> ^(FUNCAO tipo_dado IDENTIFICADOR decl_parametros? corpo FIM_FUNCAO)
    ;

procedimento
    : PROCEDIMENTO IDENTIFICADOR '(' decl_parametros? ')'
      corpo 
      FIM_PROCEDIMENTO 
      
      -> ^(PROCEDIMENTO IDENTIFICADOR decl_parametros? corpo FIM_PROCEDIMENTO)
    ;

principal
    : 'principal'
      corpo 
      'fimPrincipal' 
      
      -> ^(PRINCIPAL corpo)
    ;

/*
  Statements
*/
corpo
    : decl_variaveis*
      mult_statements? 
      
      -> ^(CORPO decl_variaveis* mult_statements?)
    ;

mult_statements
    : statements+ -> ^(STATEMENTS statements+)
    ;

statements
    : facaStm                 // do..while
    | enquantoStm             // while
    | paraStm                 // for
    | seStm                   // if
    | escolhaStm              // switch
    | statementsSemiColon ';'! 
    ;

seStm
    : 'se' '(' expressao ')' 'entao'
      mult_statements?
      senaoStm?
      'fimSe' 
      
      -> ^(SE expressao mult_statements? senaoStm?)
    ;
    
senaoStm
    : ( 'senaoSe' '(' expressao ')' 'entao'
      mult_statements? senaoStm? )=>
      'senaoSe' '(' expressao ')' 'entao'
      mult_statements? senaoStm?
      
      -> ^(SENAO_SE expressao mult_statements?) senaoStm?
    
    | 'senao' 
      mult_statements?
    
      -> ^(SENAO mult_statements)?
    ;
    
escolhaStm
    : 'escolha' '(' IDENTIFICADOR ')'
      caso+
      outrocaso?
      'fimEscolha'
      
       -> ^(ESCOLHA IDENTIFICADOR ^(CASOS caso+) outrocaso?)
    ; 
    
caso
    : 'caso' (INTEIRO_LITERAL | CHAR_LITERAL)
      mult_statements?
      
      -> ^(CASO INTEIRO_LITERAL? CHAR_LITERAL? mult_statements?)
    ;
    
outrocaso
    : 'outroCaso'
      mult_statements?
      
      -> ^(OUTROCASO mult_statements?)
    ;
           
paraStm
    : 'para' '(' IDENTIFICADOR 'de' expressao 'ate' expressao 'passo' expressao ')' 'faca'
      mult_statements?
      'fimPara' 
      
      -> ^(PARA IDENTIFICADOR expressao expressao expressao mult_statements?)
    ;
    
enquantoStm
    : 'enquanto' '(' expressao ')' 'faca'
      mult_statements?
      'fimEnquanto' 
      
      -> ^(ENQUANTO_STM expressao mult_statements?) 
    ;
    
facaStm
    : FACA corpo_facaStm
    
      -> ^(FACA corpo_facaStm)
    ;

corpo_facaStm
    : ('enquanto' '(' expressao ')' ';' )=> 'enquanto' '(' expressao ')' ';'
      
      -> ^(ENQUANTO expressao)
    
    | statements corpo_facaStm
    ;

statementsSemiColon
    : atribuicaoStm
    | limpatela    
    | chamadaRotina
    | lerStatement 
    | lerCaracterStatement
    | escreverStatement 
    | interrompa        
    | retornaStm        
    ;
    
atribuicaoStm
    : IDENTIFICADOR ('[' identVetor=expressao ']' ('[' identMatriz=expressao ']')?)? '=' valorAtribuido=expressao 
      -> ^(ATRIBUICAO IDENTIFICADOR $valorAtribuido $identVetor? $identMatriz?)
    ;
    
limpatela
    : LIMPA_TELA '(' ')' 
      -> ^(LIMPA_TELA)
    ;
    
chamadaRotina
    : IDENTIFICADOR '(' parametros_chamada? ')' 
      -> ^(CHAMADA IDENTIFICADOR parametros_chamada?)
    ;

lerStatement
    : LEIA '(' expressao ')' 
      -> ^(LEIA expressao)
    ;
    
lerCaracterStatement
    : LEIA_CARACTER '(' expressao ')'
      -> ^(LEIA_CARACTER expressao)
    ;
    
escreverStatement
    : 'escreva' '(' expressaoEscreva (',' expressaoEscreva)* ')' 
      -> ^(ESCREVA expressaoEscreva expressaoEscreva*)
    | 'escreval' '(' expressaoEscreva (',' expressaoEscreva)* ')' 
      -> ^(ESCREVAL expressaoEscreva expressaoEscreva*)
    ;
    
expressaoEscreva
    : (expressao (','|')'))=> expressao 
    | (expressao ':' INTEIRO_LITERAL)=> expressao ':' total=INTEIRO_LITERAL (':' decimal=INTEIRO_LITERAL)?
      -> ^(FORMATA_REAL expressao ^(FORMATA_REAL_TOTAL $total) ^(FORMATA_REAL_DECIMAL $decimal?))
    | expressao ':' ':' decimal=INTEIRO_LITERAL
      -> ^(FORMATA_REAL expressao ^(FORMATA_REAL_TOTAL) ^(FORMATA_REAL_DECIMAL $decimal?))
    ;
    
interrompa
    : INTERROMPA
    ;

retornaStm
    : 'retorna' expressao
      -> ^(RETORNA expressao)
    ;   

/*
  Expressoes
*/
termo
    :   IDENTIFICADOR
    |   IDENTIFICADOR '[' expressao ']' ('[' expressao ']')? -> ^(TERMO_VETOR IDENTIFICADOR expressao expressao?)
    |   '('! expressao ')'!
    |   chamadaRotina
    |   TEXTO_LITERAL
    |   CHAR_LITERAL
    |   BOOL_LITERAL
    |   INTEIRO_LITERAL
    |   REAL_LITERAL
    |   funcoesEmbutidas
    ;
    
formataReal
    : '{' expressao '}' ':' total=INTEIRO_LITERAL? ':' decimal=INTEIRO_LITERAL?
    
      -> ^(FORMATA_REAL expressao ^(FORMATA_REAL_TOTAL $total?) ^(FORMATA_REAL_DECIMAL $decimal?))
    ;
    
funcoesEmbutidas
    :   'comparaTexto' '(' expressao ',' expressao ')' -> ^(COMPARA_TEXTO expressao expressao)
    |   'tamanhoTexto' '(' expressao ')' -> ^(TAMANHO_TEXTO expressao)
    |   'abs' '(' expressao ')' -> ^(ABS expressao)
    |   'exp' '(' expressao ',' expressao ')' -> ^(EXP expressao expressao)
    |   'pi' '(' ')' -> ^(PI)
    |   'raizQuadrada' '(' expressao ')' -> ^(RQUAD expressao)
    |   'aleatorio' '(' expressao ')' -> ^(ALEAT expressao)
    |   'maiusculo' '(' expressao ')' -> ^(MAIUSC expressao) 
    |   'minusculo' '(' expressao ')' -> ^(MINUSC expressao)
    |   'copia' '(' expressao ',' expressao ',' expressao ')' -> ^(COPIA expressao expressao expressao)
    |   'maiusculoCaracter' '(' expressao ')' -> ^(MAIUSC_CHAR expressao)
    |   'minusculoCaracter' '(' expressao ')' -> ^(MINUSC_CHAR expressao)
    |   'asciiCaracter' '(' expressao ')' -> ^(ASCII_CHAR expressao)
    |   'caracterAscii' '(' expressao ')' -> ^(CHAR_ASCII expressao)
    ;

negacao
    :   NEGACAO_LOGICA^* termo
    ;
    
unario
    :   ('+'! | negacao_matematica^)* negacao
    ;
    
negacao_matematica
    : '-' -> NEGACAO_MATEMATICA
    ;
   
mult
    :   unario ((MULTIPLICACAO^ | DIVISAO^ | DIVISAO_INT^ | MOD^) unario)*
    ;
    
soma
    :   mult ((ADICAO^ | SUBTRACAO^) mult)*
    ;

comparacao
    :   soma ((IGUAL^ | DIFERENTE^ | MENOR^ | MENOR_IGUAL^ | MAIOR_IGUAL^ | MAIOR^) soma)*   
    ;
    
expressao
    :   comparacao ((E^ | OU^) comparacao)*
    ;


/* 
  Parametros e variaveis 
*/
tipo_dado
    : TIPO_INTEIRO
    | TIPO_REAL
    | TIPO_CARACTER
    | TIPO_TEXTO
    | TIPO_LOGICO
    ; 

decl_variaveis
    :  decl_variavel^ ';'!
    ;   

decl_variavel
    : tipoDado=tipo_dado identificadores 
    
      -> ^(DECL_VARIAVEL tipo_dado identificadores)
	  ;
	  
identificadores
    : IDENTIFICADOR (('[' vetor=INTEIRO_LITERAL ']') | ('[' vetor=INTEIRO_LITERAL ']' '[' matriz=INTEIRO_LITERAL ']'))? (',' identificadores)?
    
      -> ^(IDENT IDENTIFICADOR $vetor? $matriz?) identificadores?
    ;

tipo_identificador
    : tipo_dado IDENTIFICADOR 
    
      -> ^(TIPO_IDENTIFICADOR tipo_dado IDENTIFICADOR)
    ;
    
decl_parametros
    : decl_parametro (',' decl_parametro)* 
    
      -> ^(PARAMETROS decl_parametro*) 
    ;

decl_parametro
    : tipo_identificador -> ^(PARAMETRO tipo_identificador)
    | REFERENCIA tipo_identificador -> ^(PARAMETRO tipo_identificador REFERENCIA)
    | tipo_identificador PARAM_VETOR -> ^(PARAMETRO tipo_identificador PARAM_VETOR)
    | tipo_identificador PARAM_MATRIZ -> ^(PARAMETRO tipo_identificador PARAM_MATRIZ)
    ;
       
parametros_chamada
    : expressao (',' expressao)* 
    
      -> ^(PARAMETROS_CHAMADA expressao *)
    ;    

/* 
  Tokens
*/
TEXTO_LITERAL 
    : '"'
      {StringBuilder string = new StringBuilder();}
      (   '"' '"'           {string.appendCodePoint('"');}
      |  c=~('"'|'\r'|'\n')  {string.appendCodePoint(c);}
      )* 
      '"' 
      {setText(string.toString());}
    ;
CHAR_LITERAL : '\'' . '\'' {setText(getText().substring(1, 2));};
BOOL_LITERAL : 'verdadeiro' | 'falso';
fragment LETRA : ('a'..'z' | 'A'..'Z'); 
fragment DIGITO : '0'..'9';
INTEIRO_LITERAL : DIGITO+ ;
REAL_LITERAL : DIGITO+ '.' DIGITO* ;
IDENTIFICADOR : LETRA(LETRA | DIGITO | '_')* ;
WHITESPACE : (' ' | '\t' | '\n' | '\r' | '\f')+ {$channel = HIDDEN;} ;
COMENTARIO : '//' ~('\n' | '\r')* {$channel = HIDDEN;} ;
COMENTARIO_MULTILINHA : '/*' .* '*/' {$channel = HIDDEN;} ;
